
(defvar wppb:preparse-usage
;;AUTODOC: wppb:preparse "preparse wool files and generate WPPB preparsed files"
"sml-preparse [options] files...

where files can be:
  * a sml file which is analyzed and preparsed (written back with the p
    extension after its own extension)
  * a dictionary definition (of the form *.wdd) which will be used to
    factorize definitions of all the symbols found in files, and will be
    written back if needed as compiled forms (*.wld)
  * a dictionary (of the form *.wld), making subsequently processed files only
    refer to its contents without having to re-define its objects in the file.
  * a directory, which is scanned and whose files and subdirs are recursively
    preparsed

options can be given Lisp style (keywords) or Unix style (- prefixed)
before files e.g sml-preparse -v file or sml-preparse :v file
  -v          verbose: prints information
  -q          quiet: do not output anything (normally only file names
  -d name     use name as dir (loads name.wld) 
  -t number   threshold: put an item into the here-dictionnary if seen at least
              number times or more in file (default 2)
  -nd         do not preparse the defun, defunq and defmacro definitions
  -nc         do not compact all equal objects in file to the same object
  -eq         to compact, use eq not equal (only atoms will be compacted)
  -x ext      add allowed extension for parsable files, eg: -x .eg -x .ed
  -D  or -V   debug:   prints LOTS of information
  -fv number  force internal version to be number (5015000 for 5.15.0)
"
)
;;=============================================================================
;;                    utilities
;;=============================================================================
;;AUTODOC: checksum-string "computes a checksum on a string"
;;(checksum-string string)
;;Computes and returns a simple 32bit checksum which is just the sum of
;;all characters + length of string

(defun checksum-string (s &aux (res (apply + (List s))))
  (if res (+ res (length s)) 0)
)

;;AUTODOC: checksum-file "computes a checksum on a file"
;;(checksum-file filename)
;;Computes and returns a simple 32bit checksum which is just the sum of
;;all characters + length of file
;;Triggers an open error if file could not be opened

(defun checksum-file (f)
  (checksum-string (String (open f)))
)

;;=============================================================================
;;                    definitions
;;=============================================================================
(setq *package* 'wppb)
(setq wppb:version 0)
(defvar wppd:excluded-items '(() t))

;; COLAS: added this line for testing...
(if (not (boundp 'sml:layers)) (setq sml:layers '(Sml 1)))

(defun wppb:init-options ()
  (setq wppd:used-list (list ()))
  (setq debug ())
  (setq verbose ())
  (setq no-defun-compilation ())
  (setq no-compaction ())
  (setq wppb:equal equal)
  (setq wppb:seek wppb:seek-equal)
  (setq wppb:get get)
  (setq wppb:put put)
  (setq wppb:quote-write ())
  (setq wppb:preprocess:name ())
  (setq wppb:here-threshold 2)
  (setq quiet ()))

(defun wppb:process-option (o argument &aux 
    (used-argument? 0)			;set to 1 if you used argument
  )
  (if
    (= "-D" o) (progn (setq debug t) (setq verbose t))
    (= "-V" o) (progn (setq debug t) (setq verbose t))
    (= "-v" o) (setq verbose t)
    (= "-nd" o) (setq no-defun-compilation t)
    (= "-nc" o) (setq no-compaction t)
    (= "-q" o) (progn (setq quiet t) (setq verbose ()) (setq debug ()))
    (= "-eq" o) (progn (setq wppb:equal eq) (setq wppb:seek seekq)
      (setq wppb:get hash-get-eq wppb:put hash-put-eq))
    (= "-d" o) (with (name (+ argument wppb:extension-wld)) (wppb:use-wld name)
      (incf used-argument?)
    )
    (= "-t" o) (progn (setq wppb:here-threshold (Int argument))
      (incf used-argument?)
    )
    (= "-x" o) (progn (lappend *load-extensions* argument)
      (incf used-argument?))
    (= "-fv" o) (progn (put sml:layers 'Sml argument) (incf used-argument?))
    (write-string wppb:preparse-usage)	;default
  )
  used-argument?
)

;;=============================================================================
;;                    toplevel
;;=============================================================================
;; toplevel applies wppb:preparse-any to all args
;; options can be unix or CL form (keywords)

(defun wppb:preparse (&rest args &aux 
    (files (copy ()))
    arg
    here-dict
    (i 0)
  )
  (wppb:init-options)
  (while (< i (length args))
    (setq arg (getn args i))
    (if (match "^[-]" arg) 
	(incf i (wppb:process-option arg (getn args (+ i 1))))
      (match "^[:]" arg)
      (incf i (wppb:process-option (+ "-" (subseq arg 1)) (getn args (+ i 1))))
      (wppb:preparse-any arg)
    )
    (incf i)
  )
)

;; main dispatcher: perform the relevant function from the file name

(defun wppb:preparse-any (file)
  (if 
    (wppb:at-end wppb:extension-wdd file) 
    (progn
      (wppb:preparse.wdd file)
      (wppb:use-wld (+ (match 
	    (+ "^(.*)[.]" (subseq wppb:extension-wdd 1) "$")
	    file 1) wppb:extension-wld))
    )

    (wppb:at-end wppb:extension-wld file)
    (wppb:use-wld file)

    (eq 'directory (file-type file)) 
    (wppb:preparse-directory file)

    (has-a-loadable-extension file) 
    (wppb:preparse-file file)

    (is-an-abbreviated-form file) 
    (wppb:preparse-any wppb:filename)

    (verbose? "Warning: file %0 not processed (unknown type)\n" file)
))

;;======================================================================= FILES
(defun wppb:preparse-file (file &aux
    (outfile (+ file wppb:extension-string))
    (dummy (verbose? "********  generating %1 from %0:\n" file outfile))
    (out (open outfile :direction :io :if-exists :supersede))
    (*quote-inlines* t)
    in
    expr
    size
    (used-dicts (list ()))
    here-dict
    dict-pos
    length-pos
    (here-item-counts (Hashtable ()))
    buf
    (*package* ())
    (wppb:todo (list))
    (version (append "@(#)Sml: " ((lambda (v)
					    (if (> v 1000000)
						(append
						 (coerce (/ v 1000000) String) 
						 "."
						 (coerce (/ (mod v 1000000)
							    1000) String) 
						 "."
						 (coerce (mod v 1000) String))
					      (append
					       (coerce (/ v 1000) String) 
					       "."
					       (coerce (mod v 1000) String))))
					  (get sml:layers 'Sml))))
    )
  (if (not no-compaction) (setq here-dict (make-here-dict file)))

  (setq in (open file))
  (wppb:writeBE32 386928642 out)	;magic
  (while (> 20 (length version))
    (nconc version " "))
  (write-string version out)
  (wppb:writeBE32 (list-get sml:layers 'Sml) out)	;SML Version
  (setq length-pos (file-position out))
  (wppb:writeBE32 0 out)		;length
  (wppb:writeBE32 0 out)		;checksum
  (wppb:writeBE8 wppb:version out)	;version
  (setq dict-pos (file-position out))
  (catch 'EOF
    (while t
      (if wppb:todo (progn (setq expr (getn wppb:todo 0)) (delete wppb:todo 0))
	(setq expr (read in))
	)
      (if (boundp 'expr)
	  (wppb:preprocess expr out)
	(wppb:preprocess-undefined out)
	)
      ))
  (wppb:write-eof out)
  (file-position out dict-pos)
  (setq buf (String out))		;saves what has been written in string
  (file-position out dict-pos)
  (dotimes (i (length used-dicts))
    (if (getn used-dicts i)
      (wppb:write-WPPDInc i out)
  ))
  (if debug (progn 
      (PV used-dicts)
      (print-format "dictionnaries used: ") 
      (dotimes (i (length used-dicts))
	(if (getn used-dicts i) (print-format "%0 " (getn wppd:used-list i)))
      )
      (if here-dict (print-format "<here-dictionary>"))
      (write-line "")
  ))
      
  (wppb:write-WPPDHere here-dict out)
  (write-string buf out)
  (setq size (file-position out))	;write size now
  (file-position out length-pos)
  (wppb:writeBE32 size out)
					;write checksum, which is the checksum of
					;file with checksum field set to 0
  (flush out)
  (if (/= *machine* 'amiga)
      (wppb:writeBE32 (checksum-file outfile) out)
    )
  (flush out)
  (if (boundp 'wppb:crypt-file)
      (wppb:crypt-file outfile)
    )
  (if (not quiet)
      (print-format 
       "%1: %0 bytes written (from %2)\n" size outfile (file-position in))
    )
  )

(defun random-number (initial)
  (setq initial (/ initial 2))
  ((lambda (x) (if (= x 0) (random-number initial) x))
   (random initial)))

(defun wppb:crypt-file (outfile &aux
    (string (coerce (open outfile :direction :input) String))
    (len (length string))
    (output (open outfile :direction :io :if-exists :supersede))
    (i 0)
    (rand 0)
    (rank 37)
  )
  (map () '(lambda (x)
      (when (= i rank)
	(setq rand (random-number len))
	(setq rank (+ rank rand 8))
	(setq i (+ i 8))
	(wppb:writeBE32 rand output)
	(wppb:writeBE32 (logxor rank rand) output))
      (write-char x output)
      (setq i (+ i 1)))
    string)
  (close output)))

;;=================================================================== DICT DEFS
(defun wppb:preparse.wdd (file &aux
    (in (open file))
    (name (+ "wppd:" (subseq file 0 (- (length file) 
	    (length wppb:extension-wdd)))))
    (outfile (+ (subseq file 0 (- (length file) (length wppb:extension-wdd))) 
	wppb:extension-wld))
    (out (open outfile :direction :io :if-exists :supersede))
    item checksum-pos size-pos check
    (count 1)				;checksum in 0
    size
    (wppd:used-list (list ()))
  )
  (verbose? "Generating dictionnary %0 from %1 : \n" outfile file)
					;header
  (wppb:writeBE32 386928642 out)	;magic
  (wppb:writeBE32 0 out)		;length
  (wppb:writeBE32 0 out)		;checksum
  (wppb:writeBE8 wppb:version out)	;version
  (write-char wppb:type:Set out)
  (wppb:write (intern name) out)
  (write-char wppb:type:ListML out)
  (setq size-pos (file-position out)) ;we will update later
  (wppb:writeBE32 0 out)
  (write-char 0 out)
  (setq checksum-pos (file-position out)) ;we will update later
  (wppb:write-Int32 0 out) 
  (catch 'EOF 
    (while t 
      (setq item (read in))
      (if (and (boundp 'item)
	       (not (seekq wppd:excluded-items item))
      ) (progn
	  (if debug (print-margin-line " -> %r0\n" item))
	  (wppb:write item out)
	)
      )
      (incf count)
  ))
  (wppb:write-eof out)
  (close in)
  (file-position out size-pos)
  (wppb:writeBE32 count out)
  (file-position out checksum-pos)
  (wppb:write-Int32 (setq check (checksum-file file)) out)
  (setq size (progn (file-position out 0 2) (file-position out)))
  (file-position out 4)
  (wppb:writeBE32 size out)
  (verbose? "dictionnary of %0 items, size %2 checksum %1\n"
    (- count 1) check size)
)

;;=================================================================== DICT USES
(defun wppb:use-wld (file &aux
    (*load-extensions* (+ (list "") *load-extensions*))
  )
  (load file)
  (lappend wppd:used-list 
    (intern (+ "wppd:" (subseq file 0 (- (length file) 
	    (length wppb:extension-wld)))))
  )
  (verbose? "Including dictionnary %0 \n" file)
)

;;=============================================================== DIR recursion
(defun wppb:preparse-directory (dir &aux
    (prefix (if (match "[/]$" dir) dir (+ dir "/")))
    (dict-def-list (copy ()))
    (dict-list (copy ()))
    (files-list (copy ()))
  )
  (verbose? "preparsing files in directory %0\n" dir)
  (dolist (file (directory dir))
    (setq file (+ prefix file))
    (if (wppb:at-end wppb:extension-wdd file) (progn 
	(lappend dict-def-list file)
	(lappend dict-list file)
      )
      (wppb:at-end wppb:extension-wld file) (lappend dict-list file)
      (lappend files-list file)
  ))
  (dolist (l (list dict-def-list dict-list files-list))
    (dolist (file l)
      (wppb:preparse-any file)
  ))
)

;;=============================================================================
;;                    Preparser
;;=============================================================================

(defun wppb:preprocess (expr stream &aux
    wppb:preprocess:tmp
    wppb:preprocess:name
    printed
  )
  (if debug (print-margin-line " -> %r0\n" expr))
  
  (if (and (= (type-of expr) List)
      (= 3 (length expr))
      (eq (getn expr 1) '*package*)
      (eq (getn expr 0) 'setq)
    ) (progn
      (eval expr)
    )
    (and (not no-defun-compilation)	;preparse function defs
      (= (type-of expr) List)
    )
    (if
      (seekq '(defun defunq defmacro) (getn expr 0)) ;lambdas
      (progn
	(wppb:extract-documentation expr (getn expr 1))
	(setq wppb:preprocess:tmp (apply lambda (subseq expr 2)))
	(if (eq 'defunq (getn expr 0)) (type-replace wppb:preprocess:tmp FExpr)
	  (eq 'defmacro (getn expr 0)) (type-replace wppb:preprocess:tmp Macro)
	)
	(setq wppb:preprocess:name (getn expr 1))
	(if verbose (setq printed t))
	(verbose? " preparsing function %0 (%1)\n"
	  wppb:preprocess:name (type-of wppb:preprocess:tmp)
	)
	(wppb:write-set wppb:preprocess:name wppb:preprocess:tmp stream)
	(setq expr ())
      )
      (eq 'defmethod (getn expr 0))	;defmethods
      (with (wppb:preprocess:class () l ())
	(setq wppb:preprocess:name (getn expr 1))
	(setq wppb:preprocess:class #[expr 2 0 1])
	(wppb:extract-documentation expr (+ wppb:preprocess:class "."
					    wppb:preprocess:name
        ))
	(setq wppb:preprocess:tmp (subseq expr 2))
	(put wppb:preprocess:tmp 0 (setq l (copy (getn wppb:preprocess:tmp 0))))
	(put l 0 #[l 0 0])
	(setq wppb:preprocess:tmp (apply lambda wppb:preprocess:tmp))
	(if verbose (setq printed t))
	(verbose? " preparsing method %0 for type %1\n"
	  wppb:preprocess:name wppb:preprocess:class
	)
	(wppb:write-eval 
	  (list 'add-method wppb:preprocess:class `',wppb:preprocess:name
	    wppb:preprocess:tmp
	  )
	  stream
	)
	(setq expr ())
      )
      
    )
  )
  (if (and verbose (not printed) (not debug)) (verbose? "."))

  (if expr (wppb:write-eval expr stream))
)

(defun wppb:preprocess-undefined (stream)
  (wppb:write-undefined stream)
)

(defun wppb:extract-documentation (expr name &aux
  )
  (if (and (> (length expr) 4) (= String (type-of (getn expr 3)))) 
    (with (doc (getn expr 3))
      (lappend wppb:todo `(documentation ',name ,doc))
      (delete expr 3)
)))


;;=============================================================================
;;                    "compilation" functions
;;=============================================================================

(defun make-here-dict (file &aux
    (here (list))
    (counts (Hashtable ()))
    (sorted-counts (list))
    expr
    all
    (local 0)
    (in (open file))
  )
  (verbose? "creating here-dictionnary: reading ")
  (catch 'EOF
    (if debug
	(with (add-here add-here-debug)
	  (while t
	    (setq expr (read in)) (verbose? "%0 " (file-position in))
	    (if (boundp 'expr) 
		(with (already-visited (vector))
		  (add-here expr counts))
	      ))
	  (verbose? "\n")
	  )
      (while t
	(setq expr (read in))
	(if (boundp 'expr) (add-here expr counts))
	)
      )
    )
  (close in)
  (setq all (length counts))
  (verbose? "(%0), finding repeats " all)
  ;;remove elts already belonging to other dicts
  (dohash (elt count counts)
    (catch 'in-dict
					;check if built in: (), t, numbers...
      (if (or (seekq wppd:excluded-items elt)
	  (and (= (type-of elt) Int) (<= elt 20) (>= elt -2))
	)
	(throw 'in-dict t)
      )
      (dolist (dict wppd:used-list)	;in imported dicts?
	(if (setq offset (wppb:seek (eval dict) elt)) 
	  (throw 'in-dict 
	    (put used-dicts (seekq wppd:used-list dict) t) 
	    t))
      )
      (incf local)
      (if (>= count wppb:here-threshold)
	(lappend sorted-counts (list elt count))
      )
  ))
  (verbose? "(%0), sorting, " (length sorted-counts))
  ;; sort by high counts first
  (sort sorted-counts (lambda (p1 p2) (- (compare #[p1 1] #[p2 1]))))
  (verbose? "done.\n")
  (if debug (with (*print-level* 10)
      (print-format "Counts for here-dictionnary: 
(%2 elements seen at least %1 times) out of %0 elts):\n" 
	local wppb:here-threshold (length sorted-counts)
      )
      (dolist (pair sorted-counts)
	(print-margin-line "    [%0] %1\n" #[pair 1] #[pair 0])
  )))
  (dolist (pair sorted-counts)
    (lappend here #[pair 0])
  )
  (if here
    (verbose? "here-dictionary: %0 items (seen at least %2 times) out of %1\n"
      (length here) all wppb:here-threshold
    )
    (verbose? "no here dict\n")
  )
  here
)

(defun add-here (obj counts &aux count)
  (if (boundp 'obj) (progn            ;obj may be set to *undefined* !
      (add-here-obj obj counts)               ;recurse on sons first
      (wppb:put counts obj (setq count (+ 1 (wppb:get counts obj 0))))
)))

(setq Errors:AlreadyVisited "already visited: %r0")

(defun add-here-debug (obj counts &aux count)
  (if (boundp 'obj) (progn            ;obj may be set to *undefined* !
      (if (seekq already-visited obj)
      (progn
        (print-format "\nERROR: infinite loop while parsing expr:\n")
        (dotimes (i (length already-visited))
          (print-format "  [%0] %r1\n" i (i already-visited))
        )
        (print-format "  [%0] %r1\n" (length already-visited) obj)
        (error Errors:AlreadyVisited obj)
      )
      (if (and obj (typep obj List)) (lappend already-visited obj))
      )
      (add-here-obj obj counts)               ;recurse on sons first
      (wppb:put counts obj (setq count (+ 1 (wppb:get counts obj 0))))
)))


(defun add-here-obj-atomic (obj counts))

(defmethod add-here-obj ((obj Atom) counts))

(defmethod add-here-obj ((obj List) counts)
  (dolist (elt obj) (add-here elt counts))
)
(defmethod add-here-obj ((obj QuotedExpr) counts)
  (add-here #[obj 1] counts)
)

(defmethod add-here-obj ((obj Expr) counts)
  (dolist (elt (List obj)) (add-here elt counts))
)
  
(if (not (boundp 'copy-method)) 
  (defun copy-method (name &rest args)
    (dohash (classfrom classto args)
      (add-method classto name (find-method classfrom name))
    )
))

(copy-method 'add-here-obj
  List Vector
  List Locator
  List ListInline
  Atom Constant
  Atom ExActive
  Atom Active
  Atom Keyword
  Atom String
  Atom Int
  Atom Real
  Expr FExpr
  Expr Macro
)
(if  (boundp 'Unsigned) (copy-method Atom Unsigned))


;;=============================================================================
;;                    THE main wppb:write function
;;=============================================================================

(defun wppb:write (obj stream &aux offset id c
  )
  (if (boundp 'obj) (progn
      (catch 'in-dict
      (dolist (dict wppd:used-list)   ;in imported dicts?
        (if (setq offset (wppb:seek (eval dict) obj))
          (throw 'in-dict
            (wppb:write-wppd-ref (setq id (seekq wppd:used-list dict))
              offset stream)
            (put used-dicts id t)
      )))
      (if (and here-dict 
                                      ;in here dict? (repeated item)
          (setq offset (wppb:seek here-dict obj)))
        (throw 'in-dict (wppb:write-wppd-ref-here offset stream))
      )
      (wppb:write-obj obj stream)
    ))
    (wppb:write-undefined stream)
  )
)


;; the same without dict reference

(defun wppb:write-nodict (obj stream)
  (wppb:write-obj obj stream)
)

;;=============================================================================
;;                    misc wppb:write functions
;;=============================================================================

(defun wppb:write-eof (stream)
  (write-char wppb:type:EOF stream)
  t
)

(defun wppb:write-eval (obj stream)
  (write-char wppb:type:Eval stream)
  (wppb:write obj stream)
  t
)

(defun wppb:write-set (var val stream)
  (write-char wppb:type:Set stream)
  (wppb:write var stream)
  (wppb:write val stream)
  t
)

;; write an int in long format
(defun wppb:write-Int32 (int stream)
  (write-char wppb:type:IntP stream)
  (wppb:writeBE32 int stream)
  t
)

(defun wppb:write-wppd-ref (d-id id stream)
  (if (or (> d-id 255) (> id 255)) (progn
      (write-char wppb:type:WPPDRefL stream)
      (wppb:writeBE16 d-id stream)
      (wppb:writeBE16 id stream)
    )
    (progn
      (write-char wppb:type:WPPDRef stream)
      (write-char d-id stream)
      (write-char id stream)
    )
  )
  t
)

(defun wppb:write-wppd-ref-here (id stream)
  (wppb:write-wppd-ref 0 id stream)
)

(defun wppb:write-WPPDInc (i stream)
  (write-char wppb:type:WPPDInclude stream)
  (wppb:writeBE16 i stream)
  (wppb:write-obj (get wppd:used-list i) stream)
  (wppb:writeBE32 (get (eval (get wppd:used-list i)) 0) stream)
  t
)

(defun wppb:write-WPPDHere (l stream)
  (if l (progn
      (write-char wppb:type:WPPDHere stream)
      (wppb:writeBE16 0 stream)
      (wppb:writeBE16 (length l) stream)
      (with (wppb:write wppb:write-nodict)
	(dolist (obj l)
	  (wppb:write obj stream)
      ))
  ))
  t
)

;;=============================================================================
;;                    misc utils
;;=============================================================================

(defun has-a-loadable-extension (file)
  (catch 'Ok
    (dolist (ext *load-extensions*)
      (if (and (/= "" ext)
        (match (+ (quote-string-for-regexp ext) "$") file)
      )
      (throw 'Ok t)
    ))
    ()
))

(defun verbose? (&rest args)
  (if verbose (progn
      (apply print-format args)
      (flush ())
)))

(defun is-an-abbreviated-form (file &aux filename
  )
  (catch 'Ok
    (dolist (ext *load-extensions*)
      (if (/= "" ext) 
	(if (file-stats (+ file ext))
	  (throw 'Ok (setq wppb:filename (+ file ext)))
    )))
    ()
  )
)

;; seek must be cautious and check that (= a b) AND (= b a)

(defun wppb:seek-equal (l e &aux (o (seek l e)))
  (if o 
    (if (= (getn l o) e)
      o
      (wppb:seek-equal-aux l e (+ o 1))
  ))  
)

(defun wppb:seek-equal-aux (l e offset &aux (o (seek l e offset)))
  (if o 
    (if (= (getn l o) e)
      o
      (wppb:seek-equal-aux l e (+ o 1))
  ))  
)

(defun wppb:at-end (ext s)
  (match (+ "[.]" (subseq ext 1) "$") s)
)

;;=============================================================================
;;                    Misc. wppb:write object methods
;;=============================================================================

(defvar wppb:key 7)
(defun wppb:crypt-string (s)
  (with (key wppb:key)
    (map String '(lambda (x) (setq key (logxor x key))) s)))

;;================= Now this is C code.... ====================================
;; 
;; (defmethod wppb:write-obj ((obj Atom) stream)
;;   (if (eq t obj)
;;     (write-char wppb:type:True stream)
;;     (progn
;;       (if
;; 	(> (length obj) 255)
;; 	(progn
;; 	  (write-char wppb:type:AtomL stream)
;; 	  (wppb:writeBE32 (length obj) stream)
;; 	)
;; 	(progn
;; 	  (write-char wppb:type:Atom stream)
;; 	  (write-char (length obj) stream)
;; 	)
;;       )
;;       (wppb:writeBE32 (*:hash (String obj)) stream)
;;       (write-chars (wppb:crypt-string (String obj)) () stream)
;;     )
;;   )
;;   t
;; )
;; 
;; (defun wppb:write-undefined (stream &aux (name "*undefined*"))
;;   (write-char wppb:type:Atom stream)
;;   (write-char (length name) stream)
;;   (wppb:writeBE32 (*:hash name) stream)
;;   (write-chars name () stream)
;;   t
;; )
;; 
;; (defmethod wppb:write-obj ((obj String) stream)
;;   (if (> (length obj) 255)
;;     (progn
;;       (write-char wppb:type:StringL stream)
;;       (wppb:writeBE32 (length obj) stream)
;;     )
;;     (progn
;;       (write-char wppb:type:String stream)
;;       (write-char (length obj) stream)
;;     )
;;   )
;;   (write-chars (wppb:crypt-string obj) () stream)
;;   t
;; )
;; 
;; (defmethod wppb:write-obj ((obj List) stream)
;;   (if (eq () obj)
;;     (write-char wppb:type:Nil stream)
;;     (progn (if 
;; 	(> (length obj) 255)
;; 	(progn
;; 	  (write-char wppb:type:ListML stream)
;; 	  (wppb:writeBE32 (length obj) stream)
;; 	  (if (= (type-of obj) Vector)
;; 	    (write-char 1 stream)
;; 	    (= (type-of obj) Locator)
;; 	    (write-char 2 stream)
;; 	    (write-char 0 stream)
;; 	  )
;; 	)
;; 	(= (type-of obj) List)
;; 	(progn
;; 	  (write-char wppb:type:List stream)
;; 	  (write-char (length obj) stream)
;; 	)
;; 	(progn
;; 	  (write-char wppb:type:ListM stream)
;; 	  (write-char (length obj) stream)
;; 	  (if (= (type-of obj) Vector)
;; 	    (write-char 1 stream)
;; 	    (= (type-of obj) Locator)
;; 	    (write-char 2 stream)
;; 	    (write-char 0 stream)
;; 	))
;;       )
;;       (dolist (elt obj)
;; 	(wppb:write elt stream)
;;   )))
;;   t
;; )
;; 
;; (defmethod wppb:write-obj ((obj ListInline) stream) ;inlines
;;   (wppb:write-eval (list! (copy obj)) stream)
;; )
;; 
;; (defmethod wppb:write-obj ((obj QuotedExpr) stream &aux
;;     (table '({intern "'"} 0 {intern "`"} 1 {intern ","} 2 {intern ",@"} 3))
;;   )
;;   (write-char wppb:type:QuotedExpr stream)
;;   (write-char (get table (get obj 0)) stream)
;;   (if (= '*undefined* (get obj 1))
;;       (wppb:write-undefined stream)
;;     (wppb:write (get obj 1) stream)
;;     )
;;   t
;; )
;; 
;; (defmethod wppb:write-obj ((obj Int) stream)
;;   (if (< obj -2)
;;     (progn
;;       (write-char wppb:type:IntN stream)
;;       (wppb:writeBE32 (- obj) stream)
;;     )
;;     (> obj 20)
;;     (progn
;;       (write-char wppb:type:IntP stream)
;;       (wppb:writeBE32 obj stream)
;;     )
;;     (progn
;;       (write-char wppb:type:IntS stream)
;;       (write-char (+ obj 2) stream)
;;     )
;;   )
;;   t
;; )
;;  
;; (defmethod wppb:write-obj ((obj Real) stream &aux
;;     (s (print-format String "%r0" obj))
;;   )
;;   (write-char wppb:type:Real stream)
;;   (write-char (length s) stream)
;;   (write-string s stream)
;;   (write-char 0 stream)
;;   t
;; )
;; 
;; (defmethod wppb:write-obj ((obj Expr) stream &aux
;;     (pos (file-position stream))
;;   )
;;   (if (not (wppb:write-expr obj stream)) (progn
;;       (verbose? "WARNING: function too big: %r0\n" 
;; 	(if wppb:preprocess:name wppb:preprocess:name obj)
;;       )
;;       (file-position stream pos)
;;       (wppb:write-eval (List obj) stream)
;;   ))
;;   t
;; )
;; 
;; (if (boundp 'Unsigned)
;;   (defmethod wppb:write-obj ((obj Unsigned) stream)
;;     (write-char wppb:type:IntU stream)
;;     (wppb:writeBE32 obj stream)
;;     t
;;   )
;; )

(copy-method 'wppb:write-obj
  List Vector
  List Locator
  Atom Constant
  Atom ExActive
  Atom Active
  Atom Keyword
  Expr FExpr
  Expr Macro
)

;; some package around file-stats for returning the type of a file

;;AUTODOC: file-type "return the type of a file"
"\
Usage: (file-type <name>)
returns an atom being either:
directory character-special block-special file symbolic-link socket fifo
or () if not found
last stats can be found in file-type:stats global var
(file-type \"/tmp\")     ==> 'directory
(file-type \"~/.emacs\") ==> 'file
"

(defun file-type (name &aux
    pos
  )
  (setq file-type:stats (file-stats name))
  (if file-type:stats (progn
      (if (setq pos (seek
	    file-type:plist
	    (logand S_IFMT (getn file-type:stats 'mode))
	))
	(getn file-type:plist (+ pos 1))
))))

(setq file-type:plist
  `(
    ,S_IFREG file
    ,S_IFDIR directory
    ,S_IFLNK symbolic-link
    ,S_IFCHR character-special
    ,S_IFBLK block-special
    ,S_IFSOCK socket
    ,S_IFIFO fifo
))
;;AUTODOC: quote-string-for-regexp "quote regular expression special chars"
;;takes a string and returns a copy with all characters having a special
;;signification for regular expressions quoted
;;chars are: ? * | + . ^ $ \ [ ] ( )
;;chars are quoted by including them in square brackets []
;;e.g.: "*foo*" ==> "[*]foo[*]"

(defun quote-string-for-regexp (s &aux
    (l (coerce s List))
    (res (list))
    c
  )
  (dolist (c l)
    (if (seek '(#\? #\* #\| #\+ #\. #\^ #\$ #\\ #\[ #\] #\( #\)) c)
      (progn
	(put res -1 #\[)
	(put res -1 c)
	(put res -1 #\])
      )
      (put res -1 c)
  ))
  (coerce res String)
)
  
;;AUTODOC: write-string "writes a string's contents on a stream"
;; (write-string string [stream])
;; writes the string on the stream ,or *standard-output* if not specified
;; the difference with just (write string stream) is that write-string is
;; impervious to the value of the *print-readably* flag
;;
;; (setq *print-readably* t)
;; (write "abc")		==> "abc"
;; (write-string "abc")		==> abc

(defun write-string (string &optional stream)
  (print-format stream "%n0" string))

(if (position "-decode" *arguments*)
    (if (boundp 'wppb:unpreparse)
	(apply wppb:unpreparse
	       (list-remove *arguments* (position "-decode" *arguments*)))
      (print "Option not available.\nBye.\n"))
  (apply wppb:preparse *arguments*))     ; exclude command name
;;(exit)
